
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import numpy as np

class WeatherVisualizer:
    """Visualization tools for weather predictions."""
    
    def __init__(self, figsize=(15, 10)):
        self.figsize = figsize
    
    def plot_prediction_comparison(self, true_state, pred_state):
        """Plot comparison between true and predicted states."""
        fig, axes = plt.subplots(1, len(true_state), figsize=self.figsize)
        if len(true_state) == 1:
            axes = [axes]
            
        for ax, (var_name, true_var) in zip(axes, true_state.items()):
            pred_var = pred_state[var_name]
            diff = pred_var - true_var
            
            # Create three panel plot
            img = ax.imshow(diff, cmap='RdBu_r')
            plt.colorbar(img, ax=ax)
            ax.set_title(f'{var_name} Difference')
            
        plt.tight_layout()
        return fig
    
    def plot_error_distribution(self, true_state, pred_state):
        """Plot error distribution for predictions."""
        fig, axes = plt.subplots(1, len(true_state), figsize=self.figsize)
        if len(true_state) == 1:
            axes = [axes]
            
        for ax, (var_name, true_var) in zip(axes, true_state.items()):
            pred_var = pred_state[var_name]
            errors = (pred_var - true_var).flatten()
            
            ax.hist(errors, bins=50, density=True)
            ax.set_title(f'{var_name} Error Distribution')
            
        plt.tight_layout()
        return fig
    
    def plot_global_forecast(self, forecast_state):
        """Plot global forecast on a map."""
        fig = plt.figure(figsize=self.figsize)
        ax = plt.axes(projection=ccrs.PlateCarree())
        
        for var_name, var_data in forecast_state.items():
            ax.contourf(var_data, transform=ccrs.PlateCarree())
            ax.coastlines()
            ax.gridlines()
            
        plt.title('Global Forecast')
        return fig




import numpy as np
import xarray as xr
from scipy.ndimage import gaussian_filter
from ..physics.atmospheric import AtmosphericPhysics

class WeatherDataProcessor:
    """Advanced weather data processing from our successful experiments."""
    
    def __init__(self, pressure_levels=None, grid_spacing=2.5):
        self.physics = AtmosphericPhysics(
            pressure_levels=pressure_levels,
            grid_spacing=grid_spacing
        )
        
    def calculate_derived_variables(self, data: xr.Dataset) -> xr.Dataset:
        """Calculate all derived meteorological variables."""
        derived = data.copy()
        
        # 1. Thermal variables
        if all(v in data.variables for v in ['temperature', 'pressure']):
            # Potential temperature
            derived['theta'] = self.physics.potential_temperature(
                data.temperature, data.pressure
            )
            
            # Static stability
            derived['stability'] = self.physics.static_stability(
                data.temperature, data.pressure
            )
            
            # Temperature advection
            if all(v in data.variables for v in ['u', 'v']):
                derived['temp_advection'] = self._calculate_advection(
                    data.temperature, data.u, data.v
                )
        
        # 2. Dynamic variables
        if all(v in data.variables for v in ['u', 'v']):
            # Vorticity
            derived['vorticity'] = self.physics.vorticity(
                data.u, data.v, data.latitude
            )
            
            # Divergence
            derived['divergence'] = self._calculate_divergence(
                data.u, data.v, data.latitude
            )
            
            # Kinetic energy
            derived['kinetic_energy'] = 0.5 * (data.u**2 + data.v**2)
        
        # 3. Moisture variables
        if 'specific_humidity' in data.variables:
            # Moisture flux
            if all(v in data.variables for v in ['u', 'v']):
                qu = data.specific_humidity * data.u
                qv = data.specific_humidity * data.v
                derived['moisture_flux_magnitude'] = np.sqrt(qu**2 + qv**2)
                
                # Moisture flux convergence
                derived['moisture_flux_convergence'] = -(
                    self._calculate_divergence(qu, qv, data.latitude)
                )
        
        return derived
    
    def extract_temporal_features(self, data: xr.Dataset, window_size: int = 24) -> xr.Dataset:
        """Extract temporal features using our successful methods."""
        features = data.copy()
        
        for var in data.data_vars:
            # Rolling statistics
            rolling = data[var].rolling(time=window_size, center=True)
            features[f"{var}_mean"] = rolling.mean()
            features[f"{var}_std"] = rolling.std()
            
            # Temporal gradients
            features[f"{var}_trend"] = self._calculate_trend(data[var])
            
            # Diurnal components if hourly data
            if 'time' in data[var].dims and data.time.dt.hour.size > 24:
                features[f"{var}_diurnal"] = self._extract_diurnal_component(data[var])
        
        return features
    
    def _calculate_advection(self, scalar, u, v):
        """Calculate horizontal advection of a scalar field."""
        dx = self.physics.dx * np.pi/180 * self.physics.earth_radius
        dy = self.physics.dy * np.pi/180 * self.physics.earth_radius
        
        dsdx = np.gradient(scalar, dx, axis=-1)
        dsdy = np.gradient(scalar, dy, axis=-2)
        
        return -(u * dsdx + v * dsdy)
    
    def _calculate_divergence(self, u, v, lat):
        """Calculate horizontal divergence."""
        dx = self.physics.dx * np.pi/180 * self.physics.earth_radius * np.cos(np.deg2rad(lat))
        dy = self.physics.dy * np.pi/180 * self.physics.earth_radius
        
        dudx = np.gradient(u, dx, axis=-1)
        dvdy = np.gradient(v, dy, axis=-2)
        
        return dudx + dvdy
    
    def _calculate_trend(self, data, window=24):
        """Calculate local trend using linear regression."""
        x = np.arange(window)
        rolling = data.rolling(time=window, center=True)
        
        def fit_slope(y):
            mask = ~np.isnan(y)
            if np.sum(mask) > window//2:
                return np.polyfit(x[mask], y[mask], 1)[0]
            return np.nan
        
        return rolling.reduce(fit_slope)
    
    def _extract_diurnal_component(self, data):
        """Extract diurnal cycle component."""
        # Group by hour and calculate mean
        hourly_mean = data.groupby('time.hour').mean()
        
        # Subtract daily mean to get diurnal anomaly
        daily_mean = hourly_mean.mean('hour')
        return hourly_mean - daily_mean




import numpy as np
import xarray as xr
from typing import List, Optional, Dict
from sklearn.preprocessing import StandardScaler

class WeatherFeatureEngineer:
    """Advanced feature engineering from our successful experiments."""
    
    def __init__(self):
        self.scalers: Dict[str, StandardScaler] = {}
        
    def create_feature_vectors(
        self,
        data: xr.Dataset,
        variables: List[str],
        spatial_features: bool = True,
        temporal_features: bool = True,
        normalize: bool = True
    ) -> np.ndarray:
        """Create comprehensive feature vectors for ML models."""
        features = []
        
        for var in variables:
            # Basic features
            var_data = data[var].values
            if normalize:
                var_data = self._normalize_variable(var, var_data)
            features.append(var_data)
            
            if spatial_features:
                # Spatial gradients
                gradients = self._compute_spatial_gradients(data[var])
                features.extend(gradients)
                
                # Spatial statistics
                stats = self._compute_spatial_statistics(data[var])
                features.extend(stats)
            
            if temporal_features:
                # Temporal features
                temp_features = self._compute_temporal_features(data[var])
                features.extend(temp_features)
        
        return np.stack(features, axis=-1)
    
    def _normalize_variable(self, var_name: str, data: np.ndarray) -> np.ndarray:
        """Normalize variables using stored scalers."""
        if var_name not in self.scalers:
            self.scalers[var_name] = StandardScaler()
            # Reshape to 2D for scaling
            shape = data.shape
            flat_data = data.reshape(-1, 1)
            self.scalers[var_name].fit(flat_data)
        
        # Scale the data
        shape = data.shape
        flat_data = data.reshape(-1, 1)
        scaled = self.scalers[var_name].transform(flat_data)
        return scaled.reshape(shape)
    
    def _compute_spatial_gradients(self, data: xr.DataArray) -> List[np.ndarray]:
        """Compute spatial gradients and higher-order features."""
        # First-order gradients
        dy, dx = np.gradient(data)
        
        # Gradient magnitude
        gradient_mag = np.sqrt(dx**2 + dy**2)
        
        # Laplacian
        laplacian = np.gradient(dx, axis=1) + np.gradient(dy, axis=0)
        
        # Curl (for vector fields)
        if 'vector_component' in data.dims:
            curl = np.gradient(data.sel(vector_component='v'), axis=1) -                    np.gradient(data.sel(vector_component='u'), axis=0)
            return [dx, dy, gradient_mag, laplacian, curl]
        
        return [dx, dy, gradient_mag, laplacian]
    
    def _compute_spatial_statistics(self, data: xr.DataArray) -> List[np.ndarray]:
        """Compute local spatial statistics."""
        # Local statistics with different window sizes
        windows = [3, 5, 7]
        stats = []
        
        for window in windows:
            # Moving average
            mean = self._moving_average(data, window)
            stats.append(mean)
            
            # Local standard deviation
            std = self._moving_std(data, window)
            stats.append(std)
            
            # Local extrema
            min_val = self._moving_min(data, window)
            max_val = self._moving_max(data, window)
            stats.extend([min_val, max_val])
        
        return stats
    
    def _compute_temporal_features(self, data: xr.DataArray) -> List[np.ndarray]:
        """Compute temporal features."""
        # Ensure we have time dimension
        if 'time' not in data.dims:
            return []
        
        features = []
        
        # Rolling statistics
        windows = [6, 12, 24]  # Different time windows
        for window in windows:
            rolling = data.rolling(time=window, center=True)
            features.extend([
                rolling.mean().values,
                rolling.std().values,
                rolling.min().values,
                rolling.max().values
            ])
        
        # Temporal derivatives
        dt = np.gradient(data, axis=data.get_axis_num('time'))
        features.append(dt)
        
        # Acceleration (second derivative)
        d2t = np.gradient(dt, axis=data.get_axis_num('time'))
        features.append(d2t)
        
        return features
    
    def _moving_average(self, data: np.ndarray, window: int) -> np.ndarray:
        """Compute moving average with given window size."""
        kernel = np.ones((window, window)) / (window * window)
        return self._convolve2d(data, kernel)
    
    def _moving_std(self, data: np.ndarray, window: int) -> np.ndarray:
        """Compute moving standard deviation."""
        mean = self._moving_average(data, window)
        mean_sq = self._moving_average(data**2, window)
        return np.sqrt(mean_sq - mean**2)
    
    def _moving_min(self, data: np.ndarray, window: int) -> np.ndarray:
        """Compute moving minimum."""
        from scipy.ndimage import minimum_filter
        return minimum_filter(data, size=window)
    
    def _moving_max(self, data: np.ndarray, window: int) -> np.ndarray:
        """Compute moving maximum."""
        from scipy.ndimage import maximum_filter
        return maximum_filter(data, size=window)
    
    def _convolve2d(self, data: np.ndarray, kernel: np.ndarray) -> np.ndarray:
        """2D convolution with edge handling."""
        from scipy.signal import convolve2d
        return convolve2d(data, kernel, mode='same', boundary='wrap')




import torch.nn as nn

class BaseWeatherModel(nn.Module):
    """Base class for weather prediction models."""
    def __init__(self):
        super().__init__()
        
    def compute_physics_loss(self, pred, target):
        """Compute physics-informed loss."""
        mass_conservation_loss = self.mass_conservation_constraint(pred)
        energy_conservation_loss = self.energy_conservation_constraint(pred)
        return mass_conservation_loss + energy_conservation_loss



import torch
from torch.utils.data import Dataset
import xarray as xr
import numpy as np
import h5py
from pathlib import Path
from typing import List, Optional, Tuple, Union, Dict
import fsspec

class WeatherDataset:
    """Dataset class for loading weather data from HDF5 files."""

    def __init__(self, data_path: str, variables: List[str]):
        self.data_path = Path(data_path)
        self.variables = variables
        self._load_data()

    def _load_data(self):
        self.data = {}
        for var in self.variables:
            file_path = self.data_path / f"{var}_train.h5"
            if file_path.exists():
                with h5py.File(file_path, "r") as f:
                    self.data[var] = np.array(f[var])

    def __len__(self):
        return len(next(iter(self.data.values())))

    def __getitem__(self, idx: int) -> Dict[str, np.ndarray]:
        return {var: data[idx] for var, data in self.data.items()}

class ERA5Dataset(Dataset):
    """Dataset class for loading ERA5 reanalysis data from WeatherBench 2."""
    
    VARIABLE_MAPPING = {
        't': 'temperature',
        'z': 'geopotential',
        'u': 'u_component_of_wind',
        'v': 'v_component_of_wind'
    }
    
    DEFAULT_URL = "gs://weatherbench2/datasets/era5/1959-2023_01_10-6h-64x32_equiangular_conservative.zarr"
    
    def __init__(
        self,
        data_path: Optional[str] = None,
        resolution: str = "64x32",
        variables: List[str] = ['t', 'z'],
        pressure_levels: List[int] = [850, 700, 500, 300, 200],
        time_slice: Union[slice, str, Tuple[str, str]] = slice('2015', '2016')
    ):
        """
        Initialize ERA5Dataset.
        
        Args:
            data_path: Optional path to zarr dataset. If None, uses default GCS path
            resolution: Spatial resolution of the data ('64x32' or '128x64')
            variables: List of variables to load (using short names: 't', 'z', 'u', 'v')
            pressure_levels: List of pressure levels in hPa to include
            time_slice: Time period to load (slice, string, or tuple of strings)
        """
        super().__init__()
        
        # Set up data path
        self.data_path = data_path or self.DEFAULT_URL
        if isinstance(time_slice, tuple):
            time_slice = slice(*time_slice)
            
        # Map short variable names to full names
        self.variables = [self.VARIABLE_MAPPING[v] for v in variables]
        self.pressure_levels = pressure_levels
        
        print(f"Loading data from: {self.data_path}")
        self._load_data(time_slice)
        
    def _load_data(self, time_slice: slice):
        """Load the dataset and select time period."""
        try:
            self.ds = xr.open_zarr(self.data_path)
            self.times = self.ds.time.sel(time=time_slice)
            print(f"Selected time period: {self.times[0].values} to {self.times[-1].values}")
            print(f"Variables: {self.variables}")
            print(f"Pressure levels: {self.pressure_levels}")
            
        except Exception as e:
            print(f"Error loading ERA5 data: {str(e)}")
            raise
    
    def __len__(self) -> int:
        """Return number of samples (time steps - 1 for input/target pairs)."""
        return len(self.times) - 1
    
    def __getitem__(self, idx: int) -> Dict[str, Union[torch.Tensor, Dict]]:
        """
        Get a single sample with current and next timestep.
        
        Returns:
            Dictionary containing:
                - 'input': Tensor of shape [n_vars, n_levels, lat, lon]
                - 'target': Tensor of shape [n_vars, n_levels, lat, lon]
                - 'metadata': Dictionary with sample metadata
        """
        t0 = self.times[idx].values
        t1 = self.times[idx + 1].values
        
        # Get data slices for both timesteps
        data_t0 = {}
        data_t1 = {}
        
        for var in self.variables:
            # Select data for specific time and pressure levels
            data_t0[var] = self.ds[var].sel(
                time=t0,
                level=self.pressure_levels
            ).values
            
            data_t1[var] = self.ds[var].sel(
                time=t1,
                level=self.pressure_levels
            ).values
        
        # Convert to tensors
        input_data = torch.tensor(np.stack([data_t0[var] for var in self.variables]))
        target_data = torch.tensor(np.stack([data_t1[var] for var in self.variables]))
        
        return {
            'input': input_data,
            'target': target_data,
            'metadata': {
                't0': t0,
                't1': t1,
                'variables': self.variables,
                'pressure_levels': self.pressure_levels
            }
        }

    @property
    def shape(self) -> Tuple[int, ...]:
        """Return the shape of the data (n_vars, n_levels, lat, lon)."""
        return (len(self.variables), len(self.pressure_levels), 
                self.ds.latitude.size, self.ds.longitude.size)



__version__ = "0.2.8"

from .data import WeatherDataset, ERA5Dataset
from .models import WeatherFlowMatch, PhysicsGuidedAttention, StochasticFlowModel, BaseWeatherModel
from .utils import FlowVisualizer, WeatherMetrics, WeatherEvaluator
from .training import FlowTrainer, compute_flow_loss



import pytest
import numpy as np
from weatherflow.utils import WeatherVisualizer

@pytest.fixture
def sample_data():
    grid_size = (32, 64)
    true_state = {
        'temperature': np.random.randn(*grid_size),
        'pressure': np.random.randn(*grid_size)
    }
    pred_state = {k: v + np.random.randn(*v.shape) * 0.1 
                  for k, v in true_state.items()}
    return true_state, pred_state

def test_visualizer_creation():
    vis = WeatherVisualizer()
    assert vis.figsize == (15, 10)

def test_prediction_comparison(sample_data):
    true_state, pred_state = sample_data
    vis = WeatherVisualizer()
    fig = vis.plot_prediction_comparison(true_state, pred_state)
    assert fig is not None

def test_error_distribution(sample_data):
    true_state, pred_state = sample_data
    vis = WeatherVisualizer()
    fig = vis.plot_error_distribution(true_state, pred_state)
    assert fig is not None

def test_global_forecast(sample_data):
    _, pred_state = sample_data
    vis = WeatherVisualizer()
    fig = vis.plot_global_forecast(pred_state)
    assert fig is not None




import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from typing import List, Tuple, Optional

class SinusoidalTimeEmbedding(nn.Module):
    """Time embedding using sinusoidal functions as in the Transformer."""
    
    def __init__(self, dim: int = 256):
        super().__init__()
        self.dim = dim
    
    def forward(self, t: torch.Tensor) -> torch.Tensor:
        device = t.device
        half_dim = self.dim // 2
        embeddings = np.log(10000) / (half_dim - 1)
        embeddings = torch.exp(torch.arange(half_dim, device=device) * -embeddings)
        embeddings = t[:, None] * embeddings[None, :]
        embeddings = torch.cat((embeddings.sin(), embeddings.cos()), dim=-1)
        return embeddings

class PhysicsGuidedLayer(nn.Module):
    """Physics-informed neural network layer."""
    
    def __init__(self, hidden_dim: int, dropout: float = 0.1):
        super().__init__()
        self.layer_norm = nn.LayerNorm(hidden_dim)
        self.dropout = nn.Dropout(dropout)
        
        # Physics-informed network
        self.physics_net = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim * 2),
            nn.GELU(),
            nn.Linear(hidden_dim * 2, hidden_dim),
            nn.Dropout(dropout)
        )
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        residual = x
        x = self.layer_norm(x)
        x = self.physics_net(x)
        return x + residual

class MultiScaleAttention(nn.Module):
    """Multi-scale attention mechanism for capturing different spatial scales."""
    
    def __init__(self, hidden_dim: int, num_heads: int = 8, dropout: float = 0.1):
        super().__init__()
        self.num_heads = num_heads
        self.hidden_dim = hidden_dim
        
        # Multi-head attention at different scales
        self.attention_layers = nn.ModuleList([
            nn.MultiheadAttention(hidden_dim, num_heads, dropout=dropout)
            for _ in range(3)  # 3 different scales
        ])
        
        # Scale-specific convolutions
        self.scale_convs = nn.ModuleList([
            nn.Conv2d(hidden_dim, hidden_dim, kernel_size=2**i, padding=2**(i-1))
            for i in range(3)
        ])
        
        self.output_projection = nn.Linear(hidden_dim * 3, hidden_dim)
        self.dropout = nn.Dropout(dropout)
        self.layer_norm = nn.LayerNorm(hidden_dim)
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        batch_size, seq_len, hidden_dim = x.shape
        outputs = []
        
        # Process each scale
        for attention, conv in zip(self.attention_layers, self.scale_convs):
            # Reshape for convolution
            h = x.view(batch_size, int(np.sqrt(seq_len)), int(np.sqrt(seq_len)), hidden_dim)
            h = h.permute(0, 3, 1, 2)  # [B, C, H, W]
            
            # Apply scale-specific convolution
            h = conv(h)
            
            # Reshape for attention
            h = h.permute(0, 2, 3, 1).reshape(batch_size, -1, hidden_dim)
            
            # Apply attention
            h, _ = attention(h, h, h)
            outputs.append(h)
        
        # Combine multi-scale features
        combined = torch.cat(outputs, dim=-1)
        output = self.output_projection(combined)
        output = self.dropout(output)
        
        return self.layer_norm(x + output)

class PhysicsGuidedAttention(nn.Module):
    """Enhanced physics-guided attention model for weather prediction."""
    
    def __init__(
        self,
        input_channels: int,
        hidden_dim: int = 256,
        num_layers: int = 4,
        num_heads: int = 8,
        dropout: float = 0.1,
        grid_size: Tuple[int, int] = (32, 64)  # lat, lon
    ):
        super().__init__()
        self.input_channels = input_channels
        self.hidden_dim = hidden_dim
        self.grid_size = grid_size
        
        # Input embedding
        self.input_embedding = nn.Sequential(
            nn.Linear(input_channels, hidden_dim),
            nn.GELU(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.Dropout(dropout)
        )
        
        # Time embedding
        self.time_embedding = SinusoidalTimeEmbedding(hidden_dim)
        
        # Multi-scale attention layers
        self.attention_layers = nn.ModuleList([
            MultiScaleAttention(hidden_dim, num_heads, dropout)
            for _ in range(num_layers)
        ])
        
        # Physics-guided layers
        self.physics_layers = nn.ModuleList([
            PhysicsGuidedLayer(hidden_dim, dropout)
            for _ in range(num_layers)
        ])
        
        # Output projection
        self.output_projection = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim * 2),
            nn.GELU(),
            nn.Linear(hidden_dim * 2, input_channels)
        )
        
        # Layer normalization
        self.layer_norm = nn.LayerNorm(hidden_dim)
    
    def forward(
        self,
        x: torch.Tensor,
        timestep: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        batch_size = x.shape[0]
        
        # Reshape input: [B, C, H, W] -> [B, H*W, C]
        x = x.permute(0, 2, 3, 1).reshape(batch_size, -1, self.input_channels)
        
        # Input embedding
        h = self.input_embedding(x)
        
        # Add time embedding if provided
        if timestep is not None:
            time_emb = self.time_embedding(timestep)
            h = h + time_emb.unsqueeze(1)
        
        # Apply attention and physics layers
        for attn, physics in zip(self.attention_layers, self.physics_layers):
            h = attn(h)
            h = physics(h)
        
        # Final layer norm
        h = self.layer_norm(h)
        
        # Output projection
        output = self.output_projection(h)
        
        # Reshape output: [B, H*W, C] -> [B, C, H, W]
        output = output.reshape(batch_size, *self.grid_size, -1).permute(0, 3, 1, 2)
        
        return output

    def compute_physics_loss(self, pred: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Compute physics-informed loss terms."""
        # Gradient-based physics constraints
        dx = torch.gradient(pred, dim=3)[0]  # Longitude gradient
        dy = torch.gradient(pred, dim=2)[0]  # Latitude gradient
        
        # Conservation of mass (divergence should be small)
        div = dx[..., 0] + dy[..., 1]  # For u,v components
        mass_loss = torch.mean(div ** 2)
        
        # Energy conservation
        energy_pred = torch.sum(pred ** 2, dim=(1, 2, 3))
        energy_target = torch.sum(target ** 2, dim=(1, 2, 3))
        energy_loss = torch.mean((energy_pred - energy_target) ** 2)
        
        return mass_loss + 0.1 * energy_loss



import pytest
import numpy as np
import xarray as xr
import pandas as pd
from weatherflow.data import WeatherDataset, ERA5Dataset
from pathlib import Path

def test_era5_dataset():
    try:
        # Test with default WeatherBench 2 URL
        dataset = ERA5Dataset()
        assert len(dataset) > 0
        
        # Test data loading
        sample = dataset[0]
        assert isinstance(sample, tuple)
        assert len(sample) == 2  # current and next state
        
        # Test with specific variables
        dataset = ERA5Dataset(variables=['z', 't'])
        assert len(dataset) > 0
        
        # Test with specific years
        dataset = ERA5Dataset(years=[2020])
        assert len(dataset) > 0
        
    except Exception as e:
        pytest.skip(f"ERA5 test failed: {str(e)}")




import torch
import torch.nn as nn
from typing import List, Dict, Optional, Union
import numpy as np
from ..models import PhysicsGuidedAttention
from ..utils import WeatherVisualizer
import matplotlib.pyplot as plt

class WeatherEnsemble:
    """Advanced ensemble methods from our successful experiments."""
    
    def __init__(
        self,
        models: List[nn.Module],
        weights: Optional[List[float]] = None,
        device: str = 'cuda' if torch.cuda.is_available() else 'cpu'
    ):
        self.models = [model.to(device) for model in models]
        self.weights = weights or [1/len(models)] * len(models)
        self.device = device
        self.visualizer = WeatherVisualizer()
    
    def predict(
        self,
        x: torch.Tensor,
        return_individual: bool = False
    ) -> Dict[str, torch.Tensor]:
        """Generate ensemble prediction with uncertainty estimates."""
        predictions = []
        
        for model in self.models:
            model.eval()
            with torch.no_grad():
                pred = model(x.to(self.device))
                predictions.append(pred.cpu())
        
        # Stack predictions
        predictions = torch.stack(predictions)
        
        # Weighted mean
        weights = torch.tensor(self.weights).view(-1, 1, 1, 1)
        mean = torch.sum(predictions * weights, dim=0)
        
        # Uncertainty estimates
        std = torch.std(predictions, dim=0)
        
        # Confidence intervals
        ci_lower = mean - 1.96 * std
        ci_upper = mean + 1.96 * std
        
        result = {
            'mean': mean,
            'std': std,
            'ci_lower': ci_lower,
            'ci_upper': ci_upper
        }
        
        if return_individual:
            result['individual_predictions'] = predictions
            
        return result
    
    def plot_ensemble_prediction(
        self,
        x: torch.Tensor,
        true_value: Optional[torch.Tensor] = None,
        variable_idx: int = 0
    ):
        """Visualize ensemble prediction with uncertainty."""
        prediction = self.predict(x, return_individual=True)
        
        fig, axs = plt.subplots(2, 2, figsize=(15, 12))
        
        # Mean prediction
        self.visualizer.plot_weather_field(
            prediction['mean'][variable_idx],
            title="Ensemble Mean",
            ax=axs[0, 0]
        )
        
        # Uncertainty (std)
        self.visualizer.plot_weather_field(
            prediction['std'][variable_idx],
            title="Prediction Uncertainty (Std)",
            ax=axs[0, 1],
            cmap='viridis'
        )
        
        # Individual predictions
        axs[1, 0].boxplot(
            prediction['individual_predictions'][:, variable_idx].numpy().reshape(-1)
        )
        axs[1, 0].set_title("Distribution of Predictions")
        
        # Error if true value provided
        if true_value is not None:
            error = (prediction['mean'] - true_value)[variable_idx]
            self.visualizer.plot_weather_field(
                error,
                title="Prediction Error",
                ax=axs[1, 1],
                cmap='RdBu_r'
            )
        
        plt.tight_layout()
        return fig, axs

class DiverseEnsemble(WeatherEnsemble):
    """Ensemble with diverse model architectures."""
    
    @classmethod
    def create(
        cls,
        input_channels: int,
        n_models: int = 5,
        base_config: Dict = None
    ):
        """Create ensemble with varied architectures."""
        base_config = base_config or {
            'hidden_dim': 256,
            'num_layers': 4,
            'num_heads': 8
        }
        
        models = []
        for i in range(n_models):
            # Vary architecture
            config = base_config.copy()
            config['hidden_dim'] += i * 32
            config['num_layers'] += i % 2
            config['num_heads'] += i % 4
            
            model = PhysicsGuidedAttention(
                input_channels=input_channels,
                **config
            )
            models.append(model)
        
        return cls(models)

class BaggingEnsemble(WeatherEnsemble):
    """Ensemble trained with bootstrapped data."""
    
    def __init__(
        self,
        base_model: nn.Module,
        n_models: int = 5,
        sample_size: float = 0.8
    ):
        self.base_model = base_model
        self.n_models = n_models
        self.sample_size = sample_size
        super().__init__([])
    
    def fit(
        self,
        train_data: torch.utils.data.Dataset,
        val_data: torch.utils.data.Dataset,
        **training_kwargs
    ):
        """Train ensemble members on bootstrap samples."""
        from torch.utils.data import DataLoader, SubsetRandomSampler
        
        self.models = []
        dataset_size = len(train_data)
        sample_size = int(dataset_size * self.sample_size)
        
        for i in range(self.n_models):
            # Create bootstrap sample
            indices = np.random.choice(
                dataset_size, size=sample_size, replace=True
            )
            sampler = SubsetRandomSampler(indices)
            train_loader = DataLoader(
                train_data, sampler=sampler,
                batch_size=training_kwargs.get('batch_size', 32)
            )
            
            # Initialize and train model
            model = self.base_model.__class__(
                *self.base_model.args,
                **self.base_model.kwargs
            )
            
            # Train model (implement your training logic here)
            self._train_model(
                model, train_loader, val_data, **training_kwargs
            )
            
            self.models.append(model)
        
        return self




import torch
import torch.nn as nn
from torch.cuda.amp import GradScaler, autocast
from typing import Dict, List, Optional, Callable
import wandb
import os
from pathlib import Path
from tqdm.auto import tqdm
import numpy as np

class WeatherTrainer:
    """Advanced training infrastructure from our successful experiments."""
    
    def __init__(
        self,
        model: nn.Module,
        optimizer: torch.optim.Optimizer,
        scheduler: Optional[torch.optim.lr_scheduler._LRScheduler] = None,
        device: str = 'cuda' if torch.cuda.is_available() else 'cpu',
        experiment_dir: Optional[str] = None,
        use_amp: bool = True,
        use_wandb: bool = False,
        project_name: str = 'weatherflow'
    ):
        self.model = model.to(device)
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.device = device
        self.use_amp = use_amp and 'cuda' in device
        self.scaler = GradScaler() if self.use_amp else None
        
        # Setup experiment tracking
        self.use_wandb = use_wandb
        if use_wandb:
            wandb.init(project=project_name)
            wandb.watch(model)
        
        # Setup experiment directory
        if experiment_dir:
            self.experiment_dir = Path(experiment_dir)
            self.experiment_dir.mkdir(parents=True, exist_ok=True)
        else:
            self.experiment_dir = None
        
        # Initialize best metrics
        self.best_val_loss = float('inf')
        self.best_epoch = 0
        
        # Training history
        self.history = {
            'train_loss': [],
            'val_loss': [],
            'learning_rates': [],
            'physics_losses': []
        }
    
    def train_epoch(
        self,
        train_loader: torch.utils.data.DataLoader,
        physics_weight: float = 0.1,
        grad_clip: Optional[float] = 1.0
    ) -> Dict[str, float]:
        """Train for one epoch with physics-informed loss."""
        self.model.train()
        total_loss = 0
        physics_loss = 0
        n_batches = len(train_loader)
        
        with tqdm(train_loader, desc="Training") as pbar:
            for batch_idx, (x, y) in enumerate(pbar):
                x, y = x.to(self.device), y.to(self.device)
                
                # Forward pass with automatic mixed precision
                with autocast(enabled=self.use_amp):
                    pred = self.model(x)
                    mse_loss = nn.functional.mse_loss(pred, y)
                    phys_loss = self.model.compute_physics_loss(pred, y)
                    loss = mse_loss + physics_weight * phys_loss
                
                # Backward pass with gradient scaling
                self.optimizer.zero_grad()
                if self.use_amp:
                    self.scaler.scale(loss).backward()
                    if grad_clip:
                        self.scaler.unscale_(self.optimizer)
                        torch.nn.utils.clip_grad_norm_(self.model.parameters(), grad_clip)
                    self.scaler.step(self.optimizer)
                    self.scaler.update()
                else:
                    loss.backward()
                    if grad_clip:
                        torch.nn.utils.clip_grad_norm_(self.model.parameters(), grad_clip)
                    self.optimizer.step()
                
                # Update metrics
                total_loss += mse_loss.item()
                physics_loss += phys_loss.item()
                
                # Update progress bar
                pbar.set_postfix({
                    'loss': f"{loss.item():.4f}",
                    'phys_loss': f"{phys_loss.item():.4f}"
                })
        
        return {
            'loss': total_loss / n_batches,
            'physics_loss': physics_loss / n_batches
        }
    
    def validate(
        self,
        val_loader: torch.utils.data.DataLoader
    ) -> Dict[str, float]:
        """Validate model with detailed metrics."""
        self.model.eval()
        total_loss = 0
        physics_loss = 0
        n_batches = len(val_loader)
        
        predictions = []
        targets = []
        
        with torch.no_grad():
            for x, y in val_loader:
                x, y = x.to(self.device), y.to(self.device)
                
                # Forward pass
                with autocast(enabled=self.use_amp):
                    pred = self.model(x)
                    mse_loss = nn.functional.mse_loss(pred, y)
                    phys_loss = self.model.compute_physics_loss(pred, y)
                
                total_loss += mse_loss.item()
                physics_loss += phys_loss.item()
                
                # Store predictions for analysis
                predictions.append(pred.cpu())
                targets.append(y.cpu())
        
        # Compute additional metrics
        predictions = torch.cat(predictions)
        targets = torch.cat(targets)
        
        rmse = torch.sqrt(torch.mean((predictions - targets) ** 2))
        mae = torch.mean(torch.abs(predictions - targets))
        
        return {
            'val_loss': total_loss / n_batches,
            'physics_loss': physics_loss / n_batches,
            'rmse': rmse.item(),
            'mae': mae.item()
        }
    
    def train(
        self,
        train_loader: torch.utils.data.DataLoader,
        val_loader: torch.utils.data.DataLoader,
        epochs: int,
        physics_weight: float = 0.1,
        grad_clip: Optional[float] = 1.0,
        patience: int = 10,
        save_best: bool = True
    ) -> Dict[str, List[float]]:
        """Full training loop with comprehensive monitoring."""
        
        # Early stopping setup
        patience_counter = 0
        
        for epoch in range(epochs):
            # Training
            train_metrics = self.train_epoch(
                train_loader,
                physics_weight=physics_weight,
                grad_clip=grad_clip
            )
            
            # Validation
            val_metrics = self.validate(val_loader)
            
            # Learning rate scheduling
            if self.scheduler is not None:
                if isinstance(self.scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
                    self.scheduler.step(val_metrics['val_loss'])
                else:
                    self.scheduler.step()
            
            # Update history
            self.history['train_loss'].append(train_metrics['loss'])
            self.history['val_loss'].append(val_metrics['val_loss'])
            self.history['physics_losses'].append(train_metrics['physics_loss'])
            self.history['learning_rates'].append(
                self.optimizer.param_groups[0]['lr']
            )
            
            # Logging
            if self.use_wandb:
                wandb.log({
                    'epoch': epoch,
                    **train_metrics,
                    **val_metrics,
                    'learning_rate': self.optimizer.param_groups[0]['lr']
                })
            
            # Save best model
            if val_metrics['val_loss'] < self.best_val_loss:
                self.best_val_loss = val_metrics['val_loss']
                self.best_epoch = epoch
                patience_counter = 0
                
                if save_best and self.experiment_dir:
                    self.save_checkpoint(
                        self.experiment_dir / 'best_model.pt',
                        epoch,
                        val_metrics
                    )
            else:
                patience_counter += 1
            
            # Early stopping
            if patience_counter >= patience:
                print(f"Early stopping at epoch {epoch}")
                break
            
            # Print epoch summary
            print(f"
Epoch {epoch}:")
            print(f"Train Loss: {train_metrics['loss']:.4f}")
            print(f"Val Loss: {val_metrics['val_loss']:.4f}")
            print(f"Physics Loss: {train_metrics['physics_loss']:.4f}")
            print(f"RMSE: {val_metrics['rmse']:.4f}")
            print(f"Learning Rate: {self.optimizer.param_groups[0]['lr']:.6f}")
        
        return self.history
    
    def save_checkpoint(
        self,
        path: Path,
        epoch: int,
        metrics: Dict[str, float]
    ) -> None:
        """Save model checkpoint with metadata."""
        checkpoint = {
            'epoch': epoch,
            'model_state_dict': self.model.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
            'metrics': metrics,
            'history': self.history
        }
        
        if self.scheduler is not None:
            checkpoint['scheduler_state_dict'] = self.scheduler.state_dict()
        
        if self.scaler is not None:
            checkpoint['scaler_state_dict'] = self.scaler.state_dict()
        
        torch.save(checkpoint, path)
    
    def load_checkpoint(self, path: Path) -> None:
        """Load model checkpoint with metadata."""
        checkpoint = torch.load(path, map_location=self.device)
        
        self.model.load_state_dict(checkpoint['model_state_dict'])
        self.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        
        if self.scheduler is not None and 'scheduler_state_dict' in checkpoint:
            self.scheduler.load_state_dict(checkpoint['scheduler_state_dict'])
        
        if self.scaler is not None and 'scaler_state_dict' in checkpoint:
            self.scaler.load_state_dict(checkpoint['scaler_state_dict'])
        
        self.history = checkpoint['history']
        self.best_val_loss = checkpoint['metrics']['val_loss']
        self.best_epoch = checkpoint['epoch']




import torch
import torch.nn as nn
import torch.nn.functional as F
import math
import numpy as np
from typing import List, Tuple, Optional

class Swish(nn.Module):
    """Swish activation function."""
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.sigmoid(x) * x

class ConvNextBlock(nn.Module):
    """ConvNext block for spatial processing."""
    def __init__(self, dim: int, layer_scale_init_value: float = 1e-6):
        super().__init__()
        self.dwconv = nn.Conv2d(dim, dim, kernel_size=7, padding=3, groups=dim)
        self.norm = nn.LayerNorm(dim, eps=1e-6)
        self.pwconv1 = nn.Linear(dim, 4 * dim)
        self.act = nn.GELU()
        self.pwconv2 = nn.Linear(4 * dim, dim)
        self.gamma = nn.Parameter(layer_scale_init_value * torch.ones((dim)), 
                                requires_grad=True) if layer_scale_init_value > 0 else None
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        residual = x
        x = self.dwconv(x)
        x = x.permute(0, 2, 3, 1)
        x = self.norm(x)
        x = self.pwconv1(x)
        x = self.act(x)
        x = self.pwconv2(x)
        if self.gamma is not None:
            x = self.gamma * x
        x = x.permute(0, 3, 1, 2)
        return x + residual

class WeatherFlowMatch(nn.Module):
    """Flow matching model for weather prediction."""
    def __init__(self, input_dim: int = 2048, hidden_dim: int = 256):
        super().__init__()
        self.input_dim = input_dim
        self.main = nn.Sequential(
            nn.Linear(input_dim + 1, hidden_dim),
            nn.GELU(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.GELU(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.GELU(),
            nn.Linear(hidden_dim, input_dim)
        )
    
    def forward(self, x: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
        B = x.shape[0]
        x_flat = x.reshape(B, -1)
        t = t.reshape(-1, 1).float()
        h = torch.cat([x_flat, t], dim=1)
        return self.main(h).reshape(x.shape)

# Keep existing PhysicsGuidedAttention class




import torch
import torch.nn as nn
from .base import BaseWeatherModel

class ProbabilisticEncoder(nn.Module):
    """Encodes weather data into probabilistic latent space."""
    def __init__(self, input_channels, latent_dim):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(input_channels, 256),
            nn.ReLU(),
            nn.Linear(256, latent_dim * 2)  # For mean and variance
        )
        
    def forward(self, x):
        h = self.net(x)
        return torch.chunk(h, 2, dim=-1)  # Split into mean and variance

class NormalizingFlow(nn.Module):
    """Normalizing flow for flexible probability distributions."""
    def __init__(self, latent_dim):
        super().__init__()
        self.flows = nn.ModuleList([
            nn.Linear(latent_dim, latent_dim) for _ in range(4)
        ])
    
    def forward(self, z_mu, z_var):
        z = z_mu + torch.randn_like(z_var) * torch.sqrt(z_var)
        for flow in self.flows:
            z = flow(z)
        return z

class StochasticFlowModel(BaseWeatherModel):
    """Stochastic flow model for uncertainty quantification."""
    def __init__(self, input_channels, latent_dim=32):
        super().__init__()
        self.encoder = ProbabilisticEncoder(input_channels, latent_dim)
        self.flow = NormalizingFlow(latent_dim)
        self.decoder = nn.Linear(latent_dim, input_channels)
    
    def forward(self, x, num_samples=10):
        z_mu, z_var = self.encoder(x)
        predictions = []
        
        for _ in range(num_samples):
            z = self.flow(z_mu, z_var)
            pred = self.decoder(z)
            predictions.append(pred)
            
        return torch.stack(predictions)



from .base import BaseWeatherModel
from .flow_matching import WeatherFlowMatch, ConvNextBlock, Swish
from .physics_guided import PhysicsGuidedAttention
from .stochastic import StochasticFlowModel

__all__ = [
    'BaseWeatherModel',
    'WeatherFlowMatch',
    'PhysicsGuidedAttention',
    'StochasticFlowModel',
    'ConvNextBlock',
    'Swish'
]




import torch
import numpy as np
import matplotlib.pyplot as plt
from typing import Dict, List, Tuple
from scipy.stats import norm
import seaborn as sns

class CalibrationAnalyzer:
    """Advanced calibration analysis from our experiments."""
    
    def analyze_calibration(
        self,
        predictions: torch.Tensor,  # Shape: [n_models, n_samples, ...]
        targets: torch.Tensor,
        n_bins: int = 10
    ) -> Dict[str, float]:
        """Analyze prediction calibration with multiple metrics."""
        mean_pred = predictions.mean(dim=0)
        std_pred = predictions.std(dim=0)
        
        # Compute z-scores
        z_scores = (targets - mean_pred) / (std_pred + 1e-8)
        
        # Expected vs actual coverage
        coverages = []
        expected_coverages = np.linspace(0.1, 0.9, 9)
        
        for coverage in expected_coverages:
            threshold = norm.ppf((1 + coverage) / 2)
            actual_coverage = (torch.abs(z_scores) <= threshold).float().mean()
            coverages.append(actual_coverage.item())
        
        # Compute calibration metrics
        calibration_error = np.mean((np.array(coverages) - expected_coverages) ** 2)
        
        # Reliability diagram data
        hist, bin_edges = np.histogram(z_scores.numpy(), bins=n_bins, range=(-3, 3))
        bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
        
        return {
            'calibration_error': calibration_error,
            'expected_coverages': expected_coverages,
            'actual_coverages': coverages,
            'z_score_hist': hist,
            'z_score_bins': bin_centers
        }
    
    def plot_calibration_diagnostics(
        self,
        predictions: torch.Tensor,
        targets: torch.Tensor
    ) -> Tuple[plt.Figure, plt.Axes]:
        """Create comprehensive calibration diagnostic plots."""
        metrics = self.analyze_calibration(predictions, targets)
        
        fig, axs = plt.subplots(2, 2, figsize=(15, 12))
        
        # Coverage plot
        axs[0, 0].plot(
            metrics['expected_coverages'],
            metrics['actual_coverages'],
            'b-', label='Actual'
        )
        axs[0, 0].plot([0, 1], [0, 1], 'k--', label='Ideal')
        axs[0, 0].set_xlabel('Expected Coverage')
        axs[0, 0].set_ylabel('Actual Coverage')
        axs[0, 0].set_title('Prediction Coverage Analysis')
        axs[0, 0].legend()
        
        # Z-score distribution
        sns.histplot(
            data=predictions.std(dim=0).numpy().flatten(),
            ax=axs[0, 1],
            bins=30
        )
        axs[0, 1].set_xlabel('Prediction Uncertainty (Std)')
        axs[0, 1].set_title('Uncertainty Distribution')
        
        # Reliability diagram
        axs[1, 0].bar(
            metrics['z_score_bins'],
            metrics['z_score_hist'] / metrics['z_score_hist'].sum(),
            width=0.2
        )
        x = np.linspace(-3, 3, 100)
        axs[1, 0].plot(x, norm.pdf(x) * 0.2, 'r-', label='Standard Normal')
        axs[1, 0].set_xlabel('Z-Score')
        axs[1, 0].set_ylabel('Frequency')
        axs[1, 0].set_title('Reliability Diagram')
        axs[1, 0].legend()
        
        # Error vs Uncertainty
        error = torch.abs(predictions.mean(dim=0) - targets)
        uncertainty = predictions.std(dim=0)
        axs[1, 1].scatter(
            uncertainty.numpy().flatten(),
            error.numpy().flatten(),
            alpha=0.1
        )
        axs[1, 1].set_xlabel('Predicted Uncertainty')
        axs[1, 1].set_ylabel('Absolute Error')
        axs[1, 1].set_title('Error vs Uncertainty')
        
        plt.tight_layout()
        return fig, axs




import torch
import torch.nn as nn
from .atmospheric import AtmosphericPhysics

class PhysicsGuidedLayer(nn.Module):
    """Physics-guided neural network layer from our experiments."""
    
    def __init__(self, input_dim, hidden_dim=128):
        super().__init__()
        self.physics = AtmosphericPhysics()
        self.encoder = nn.Sequential(
            nn.Linear(input_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, hidden_dim)
        )
        self.physics_projection = nn.Linear(hidden_dim, input_dim)
        
    def forward(self, x, lat=None):
        features = self.encoder(x)
        if lat is not None:
            if hasattr(x, 'temperature') and hasattr(x, 'pressure'):
                stability = self.physics.static_stability(x.temperature, x.pressure)
                features = features * torch.sigmoid(stability)
            if hasattr(x, 'u') and hasattr(x, 'v'):
                vorticity = self.physics.vorticity(x.u, x.v, lat)
                features = features + vorticity
        return self.physics_projection(features)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="weatherflow",
    version="0.2.2",  # Explicitly set to 0.2.1
    author="Eduardo Siman",
    author_email="esiman@msn.com",
    description="A Deep Learning Library for Weather Prediction",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/monksealseal/weatherflow",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
    ],
    python_requires=">=3.8",
    install_requires=[
        "torch>=1.9.0",
        "numpy>=1.20.0",
        "xarray>=0.19.0",
        "matplotlib>=3.4.0",
        "cartopy>=0.20.0",
        "wandb>=0.12.0",
        "tqdm>=4.60.0",
        "fsspec>=2021.6.0",
        "gcsfs>=2021.6.0",
        "scipy>=1.7.0",
        "netCDF4>=1.5.0",
        "h5py>=3.0.0",
    ]
)



import pytest
import torch
import numpy as np
from weatherflow.models import PhysicsGuidedAttention, StochasticFlowModel

def test_physics_guided_attention():
    model = PhysicsGuidedAttention(channels=1)
    batch_size = 4
    x = torch.randn(batch_size, 1, 32, 64)
    
    # Test forward pass
    output = model(x)
    assert output.shape == x.shape
    
    # Test physics constraints - normalize the input and output for better comparison
    with torch.no_grad():
        x_norm = x / torch.norm(x)
        output_norm = output / torch.norm(output)
        energy_before = torch.sum(x_norm ** 2)
        energy_after = torch.sum(output_norm ** 2)
        rel_diff = abs(energy_after - energy_before) / energy_before
        assert rel_diff < 0.1, f"Energy not conserved: relative difference = {rel_diff}"

def test_stochastic_flow():
    model = StochasticFlowModel(channels=1)
    batch_size = 4
    x = torch.randn(batch_size, 1, 32, 64)
    t = torch.rand(batch_size)
    
    # Test forward pass
    output = model(x, t)
    assert output.shape == x.shape
